const assert = require('assert');
const Logger = require('bunyan');
const Client = require('../lib/client/client');
const tls = require('./tls');


const DEFAULT_LOG = new Logger({
  name: 'ldaptive',
  component: 'client',
  stream: process.stderr,
  level: process.env.BUNYAN_LOG_LEVEL || 'info',
  serializers: Logger.stdSerializers
});


function Ldaptive(config) {
  assert.ok(config.url, '`url` is required');
  assert.ok(config.baseDN, '`baseDN` is required');
  assert.ok(config.manager, '`manager` is required');
  assert.ok(config.password, '`password` is required');

  this.config = {
    url: config.url,
    baseDN: config.baseDN,
    manager: config.manager,
    password: config.password,
    attributes: config.attributes || ['cn'],
    alias: config.alias || {},
    activeDirectory: config.activeDirectory || false,
    log: config.logger || DEFAULT_LOG,
  };

  // https://github.com/auth0/ad-ldap-connector/commit/1f4dd2be6ed93dda591dd31ed5483a9b452a8d2a
  // https://github.com/nodejs/node/issues/11771
  // https://github.com/auth0/ad-ldap-connector/blob/master/lib/ldap.js
  const tlsOptions = {};
  if (this.config.url.toLowerCase().substr(0, 5) === 'ldaps') {
    tlsOptions.checkServerIdentity= tls.checkServerIdentity;
  }
  this.config.tlsOptions = tlsOptions;
}


Ldaptive.prototype._createClient = function(username, password) {
  const config = this.config;
  const client = new Client(config);

  return new Promise(function (resolve, reject) {
    if (username && password) {
      let bindDN;
      if (username.indexOf('=') > 0) {
        bindDN = `${username},${config.baseDN}`;
      } else {
        bindDN = `uid=${username},${config.baseDN}`;
      }
      client.bind(bindDN, password, function (err) {
        if (err) {
          client.unbind();
          reject(err);
        } else {
          resolve(client);
        }
      });
    } else {
      resolve(client);
    }
  });
};


Ldaptive.prototype._search = function(client, filter) {
  const config = this.config;

  const options = {
    filter: filter,
    scope: 'sub',
    attributes: config.attributes,
  };

  return new Promise(function (resolve, reject) {
    const entries = [];

    client.search(config.baseDN, options, function(err, res) {
      assert.ifError(err);

      res.on('searchEntry', function(entry) {
        // console.log('searchEntry: ' + JSON.stringify(entry.object));
        entries.push(entry.object);
      });

      res.on('searchReference', function(referral) {
        console.log('referral: ' + referral.uris.join());
      });

      res.on('error', function(err) {
        reject(err);
      });

      res.on('end', function(result) {
        client.log.debug('status: ' + result.status);
        // https://github.com/mcavage/node-ldapjs/issues/382
        client.destroy();
        resolve(entries);
      });
    });
  });
};


Ldaptive.prototype.search = function(filter) {
  const config = this.config;
  const that = this;

  return this._createClient(config.manager, config.password)
    .then(function (client) {
      return that._search(client, filter);
    });
};


Ldaptive.prototype.getByUid = function(uid) {
  const config = this.config;
  const alias = config.alias || {};
  const uidAttr = alias.uid || 'uid';

  const that = this;
  return new Promise(function (resolve, reject) {
    that.search(`(${uidAttr}=${uid})`).then(function (entries) {
      if (entries.length === 1) {
        resolve(entries[0]);
      } else {
        resolve(null);
      }
    }).catch(reject);
  });
};


Ldaptive.prototype._authenticate = function(uid, password, dn) {
  const config = this.config;
  const alias = config.alias || {};
  const bindDN = dn ? dn : `uid=${uid},${config.baseDN}`;

  const that = this;

  return this._createClient().then(function (client) {
    return new Promise(function (resolve, reject) {
      client.bind(bindDN, password, function(err, res) {
        if (err) {
          if (err.name === 'InvalidCredentialsError') {
            reject(err);
          } else {
            reject(err);
          }
        } else {
          that.getByUid(uid).then(function (entry) {
            resolve(entry);
          }).catch(reject);
        }
        client.unbind();
      });
    });
  });
};


Ldaptive.prototype.authenticate = function(uid, password) {
  const config = this.config;

  const that = this;

  if (config.activeDirectory) {
    return this.getByUid(uid)
      .then(function (user) {
        return that._authenticate(uid, password, user.dn);
      })
      .catch(function (err) {
        console.log(err);
      });
  } else {
    return this._authenticate(uid, password);
  }
};



module.exports = Ldaptive;
